@@ -1,3 +1,4 @@
+import os
 import logging
 
 logger = logging.getLogger(__name__)
@@ -63,6 +64,10 @@ _configure_system()
 # Delete configuration function.
 del _configure_system
 
+# Replaced with the current commit when building the wheels.
+__commit__ = "{{RAY_COMMIT_SHA}}"
+__version__ = "1.5.1"
+
 import ray._raylet  # noqa: E402
 
 from ray._raylet import (  # noqa: E402
@@ -91,10 +96,6 @@ from ray import util  # noqa: E402
 # We import ClientBuilder so that modules can inherit from `ray.ClientBuilder`.
 from ray.client_builder import client, ClientBuilder  # noqa: E402
 
-# Replaced with the current commit when building the wheels.
-__commit__ = "{{RAY_COMMIT_SHA}}"
-__version__ = "1.4.0"
-
 __all__ = [
     "__version__",
     "_config",
@@ -144,6 +145,10 @@ __all__ += [
     "PlacementGroupID",
 ]
 
+if "RAY_EXPERIMENTAL_DATA_API" in os.environ:
+    from ray.experimental import data
+    __all__.append(data)
+
 
 # Remove modules from top-level ray
 def _ray_user_setup_function():
@@ -164,5 +169,6 @@ def _ray_user_setup_function():
 
 _ray_user_setup_function()
 
+del os
 del logging
 del _ray_user_setup_function
